"""
This part of the package contains functions that are used when working with the OmegaMHN.
This includes score and gradient functions as well as adapted penalty functions.
"""
from . import with_ssr
